from eznf import modeler
import argparse

argparser = argparse.ArgumentParser()
argparser.add_argument("-c", "--cubes", required=True, type=str, help="Cube file")
argparser.add_argument("-s", "--symmetry", required=True, type=str, help="Symmetry breaking clause file")

args = argparser.parse_args()
cube_file = args.cubes
symmetry_breaking_file = args.symmetry

encoding = modeler.Modeler()

def itokenize(line):
    if line[:-1] == '\n':
        line = line[:-1]
    tokens = line.split(' ')
    ans = []
    for t in tokens:
        try:
            t = int(t)
            if t == 0: continue
            ans.append(t)
        except:
            continue
    return ans
max_var = 0
with open(cube_file, 'r') as cb_file:
    for line in cb_file.readlines():
        tkns = itokenize(line)
        assert len(tkns) > 0
        encoding.add_clause([-t for t in tkns])
        max_var = max(max_var, max([abs(t) for t in tkns]))

with open(symmetry_breaking_file, 'r') as sb_file:
    for line in sb_file.readlines():
        tkns = itokenize(line)
        if len(tkns) == 0:
            print(line)
            
        encoding.add_clause(tkns)

print(encoding.n_vars())
encoding.serialize("cube_tautology.cnf")



import modeler


def encode(clues):
    """
    Encode a sudoku puzzle as a SAT problem.

    Parameters
    ----------
    clues : list of lists of ints
        9x9 int matrix, non-zero values represent clues

    Returns
    -------
    model : eznf.Model
        A SAT model representing the sudoku puzzle.
    """
    Z = modeler.Modeler()  # modeler object from my library, always starts like this.

    # create vars
    for i in range(9):
        for j in range(9):
            for n in range(1, 10):
                Z.add_var(f"x_{i, j, n}", description=f"Cell ({i}, {j}) gets number {n}")

    # exactly one number per cell
    for i in range(9):
        for j in range(9):
            Z.exactly_one([f"x_{i, j, n}" for n in range(1, 10)])

    # respect clues
    for i in range(9):
        for j in range(9):
            if clues[i][j] != 0:
                Z.constraint(f"x_{i, j, clues[i][j]}")

    # exactly-one constraints
    for n in range(1, 10):
        # rows
        for i in range(9):
            Z.exactly_one([f"x_{i, j, n}" for j in range(9)])

        # cols 
        for j in range(9):
            Z.exactly_one([f"x_{i, j, n}" for i in range(9)])

        # sub_grids
#        sub_grids = [[[] for sj in range(3)] for si in range(3)]
 #       for i in range(9):
  #          for j in range(9):
   #             sub_grids[i//3][j//3].append((i, j))
   #     for si in range(3):
    #        for sj in range(3):
     #           Z.exactly_one([f"x_{*cell, n}" for cell in sub_grids[si][sj]])

    return Z


sudoku_example = [[5, 3, 0, 0, 7, 0, 0, 0, 0],
                  [6, 0, 0, 1, 9, 5, 0, 0, 0],
                  [0, 9, 8, 0, 0, 0, 0, 6, 0],
                  [8, 0, 0, 0, 6, 0, 0, 0, 3],
                  [4, 0, 0, 8, 0, 3, 0, 0, 1],
                  [7, 0, 0, 0, 2, 0, 0, 0, 6],
                  [0, 6, 0, 0, 0, 0, 2, 8, 0],
                  [0, 0, 0, 4, 1, 9, 0, 0, 5],
                  [0, 0, 0, 0, 8, 0, 0, 7, 9]]

encoding = encode(sudoku_example)
# encoding.serialize("sudoku.cnf")

def sudoku_decoder(model):
    matrix = [[0 for j in range(9)] for i in range(9)]
#    print(model)
    true_vars = 0
    for i in range(9):
        for j in range(9):
            for n in range(1, 10):
                if model[f"x_{i, j, n}"]:
                    true_vars += 1
                    print(f"x_{i, j, n}")
                    matrix[i][j] = str(n)
    print("true vars: ", true_vars)
    for row in matrix:
        print(row)

encoding.solve_and_decode(sudoku_decoder)



def f_1(n):
    if n <= 4:
        return (n*(n-1)) // 2
    return f_1(n//2 + 1) + f_1((n+1)//2 + 1)

def g_1(n):
    if n <= 4:
        return 0
    return 1 + g_1(n//2 + 1) + g_1((n+1)//2 + 1)

def f_2(n):
    if n <= 4:
        return (n*(n-1)) // 2 
    return f_2(4) + f_2(n-2)

def g_2(n):
    if n <= 4:
        return 0
    return 1 + g_2(n-2)

D_Bin = {}
D_Chunks = {}
for n in range(10, 120):
    D_Bin[n] = g_1(n) 
    D_Chunks[n] = g_2(n) 

print("".join(map(str, D_Bin.items())))
print("\n\n\n")
print("".join(map(str, D_Chunks.items())))





import modeler
DATA = {}

for N in range(11, 13):
    print("Processing N = ", N)
    DATA[N] = {}
    D = {"3-chunks": modeler.Modeler(),
         "3-chunks-o": modeler.Modeler(),
         "bin-tree": modeler.Modeler()}

    for k in D:
        print("processing k = ", k)
        Z = D[k]
        for i in range(N+1):
            for j in range(N):
                Z.add_var(f"x_{i, j}", description=f"Element {i} maps to {j}")

        # at least one per i
        for i in range(N+1):
            Z.add_clause([f"x_{i, j}" for j in range(N)])

        # at most one per j
        for j in range(N):
            Z.at_most_one([f"x_{i, j}" for i in range(N+1)], type=k)

        DATA[N][k] = {"vars": Z.n_vars(),
                      "clauses": Z.n_clauses()}

        proof, time = Z.solve_with_proof()
        DATA[N][k]["proof_length"] = len(proof)
        DATA[N][k]["time"] = time / 1e9

for N_key in DATA.keys():
    print(f"N = {N_key}")
    for d_k in DATA[N_key].keys():
        print(f"    Enc = {d_k}. Results = {DATA[N_key][d_k]}")
    print("-"*20)



from setuptools import find_packages, setup

setup(
    name='eznf',
    packages=find_packages(include=['eznf']),
    version='0.48.2',
    description='A library for easily encoding problems into SAT',
    author='Bernardo Subercaseaux',
    install_requires=[],
    setup_requires=['pytest-runner'],
    tests_require=['pytest'],
    test_suite='tests',
)
